from collections import defaultdict

from django.test import TestCase
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth import get_user_model
from django.views import View
from django.test.client import RequestFactory
from django.conf import settings

from accrete.tenant import set_member, get_member, get_tenant, per_tenant
from accrete.mixins import TenantRequiredMixin
from accrete.models import Tenant, Member

User = get_user_model()


class TenantViewTestCase(TestCase):

    class TestView(TenantRequiredMixin, View):

        pass

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User(email='user@accrete.test', username='Test User')
        self.user.save()
        self.user_without_tenant = User(
            email='test@test.com', username='Test User'
        )
        self.user_without_tenant.save()

    def test_not_logged_in(self):
        request = self.factory.get('/')
        request.user = AnonymousUser()
        view = self.TestView.as_view()
        res = view(request)
        self.assertEqual(res.status_code, 302)

    def test_no_tenant(self):
        request = self.factory.get('/')
        request.user = self.user_without_tenant
        request.tenant = None
        view = self.TestView.as_view()
        res = view(request)
        self.assertTrue(
            res.url.startswith(settings.ACCRETE_TENANT_NOT_SET_URL)
        )

    def test_set_member(self):
        t1 = Tenant(name='Test 1')
        t1.save()
        m1 = Member(tenant=t1, user=self.user)
        m1.save()
        set_member(m1)
        self.assertEqual(get_member(), m1)
        self.assertEqual(get_tenant(), t1)
        set_member(None)
        self.assertEqual(get_member(), None)
        self.assertEqual(get_tenant(), None)

    def test_per_tenant(self):

        @per_tenant()
        def func_per_tenant():
            calls_per_tenant[get_tenant().pk] += 1

        t1 = Tenant(name='Test 1')
        t2 = Tenant(name='Test 1')
        t1.save()
        t2.save()
        calls_per_tenant = defaultdict(int)
        func_per_tenant()
        self.assertDictEqual(calls_per_tenant, {t1.pk: 1,t2.pk: 1})



from functools import wraps
from typing import Callable
from django.http import HttpRequest
from accrete.decorator import tenant_required as tenant_required_base
from django.http import HttpResponse
from .helper import (
    is_htmx,
    get_group_data,
    access_denied_modal_response,
    access_denied_page_response
)


def tenant_required(
    tenant_groups: list[str | tuple[str, ...]] = None,
    member_groups: list[str | tuple[str, ...]] = None,
    group_missing_action: str | Callable[
        [HttpRequest, list[str | tuple[str]], list[str | tuple[str]]],
        HttpResponse
    ] = None,
    staff_bypass: bool = True,
    redirect_field_name: str = None,
    login_url: str = None
):

    def decorator(f):

        def handle_group_missing(request, tenant_groups, member_groups):
            tenant_groups = tenant_groups or []
            member_groups = member_groups or []
            if is_htmx(request):
                return access_denied_modal_response(
                    request, get_group_data(tenant_groups, member_groups)
                )
            return access_denied_page_response(
                request, get_group_data(tenant_groups, member_groups)
            )

        @wraps(f)
        @tenant_required_base(
            tenant_groups=tenant_groups,
            member_groups=member_groups,
            group_missing_action=group_missing_action or handle_group_missing,
            staff_bypass=staff_bypass,
            redirect_field_name=redirect_field_name,
            login_url=login_url
        )
        def _wrapped_view(request, *args, **kwargs):
            return f(request, *args, **kwargs)
        return _wrapped_view
    return decorator



from functools import wraps
from typing import Callable
from django.shortcuts import redirect
from accrete import config
from django.http import HttpResponse, HttpResponseForbidden, HttpRequest
from accrete.tenant import tenant_has_group, member_has_group


def tenant_required(
    tenant_groups: list[str | tuple[str, ...]] = None,
    member_groups: list[str | tuple[str, ...]] = None,
    group_missing_action: str | Callable[
        [HttpRequest, list[str | tuple[str]], list[str | tuple[str]]],
        HttpResponse
    ] = None,
    staff_bypass: bool = True,
    redirect_field_name: str = None,
    login_url: str = None
):
    from django.contrib.auth.views import login_required

    def decorator(f):
        @wraps(f)
        @login_required(
            redirect_field_name=redirect_field_name,
            login_url=login_url
        )
        def _wrapped_view(request, *args, **kwargs):

            def handle_group_missing():
                if callable(group_missing_action):
                    return group_missing_action(
                        request, tenant_groups, member_groups
                    )
                return (
                    redirect(config.ACCRETE_GROUP_NOT_SET_URL)
                    if config.ACCRETE_GROUP_NOT_SET_URL
                    else HttpResponseForbidden()
                )

            tenant = request.tenant
            if not tenant:
                return redirect(config.ACCRETE_TENANT_NOT_SET_URL)
            if staff_bypass and request.user and request.user.is_staff:
                return f(request, *args, **kwargs)
            for tenant_group in (tenant_groups or []):
                if isinstance(tenant_group, tuple) and all([
                    tenant_has_group(g) for g in tenant_group
                ]):
                    break
                elif isinstance(tenant_group, str) and tenant_has_group(tenant_group):
                    break
                return handle_group_missing()
            for member_group in (member_groups or []):
                if isinstance(member_group, tuple) and all([
                    member_has_group(g) for g in member_group
                ]):
                    break
                elif isinstance(member_group, str) and member_has_group(member_group):
                    break
                return handle_group_missing()
            return f(request, *args, **kwargs)
        return _wrapped_view
    return decorator
